defmodule XmerlRecs do
  @moduledoc """
  Define Xmerl records using record definitions extracted from Erlang Xmerl.
  """

  require Record
  Record.defrecord(:xmlElement, Record.extract(:xmlElement,
    from_lib: "xmerl/include/xmerl.hrl"))
  Record.defrecord(:xmlText, Record.extract(:xmlText,
    from_lib: "xmerl/include/xmerl.hrl"))
  Record.defrecord(:xmlAttribute, Record.extract(:xmlAttribute,
    from_lib: "xmerl/include/xmerl.hrl"))
  Record.defrecord(:xmlNamespace, Record.extract(:xmlNamespace,
    from_lib: "xmerl/include/xmerl.hrl"))

end

defmodule Test16 do
  @moduledoc """
  Provide some examples of processing XML with SweetXml and Xmerl.
  """

  require XmerlRecs
  import SweetXml

  @doc """
  Start and return a list of processes that will handle query requests.
  """
  @spec start(Tuple.t(), integer()) :: {:ok, pid}
  def start(tree, count \\ 1) do
    pids = Enum.map(1..count, fn (_idx) ->
      pid = spawn(Test16, :query_proc, [tree])
      pid
    end)
    {:ok, pids}
  end

  @doc """
  Handle query requests.
  """
  @spec rpc(pid, atom, String.t()) :: {:ok, any} | {:error, String.t()}
  def rpc(pid, command, pattern) do
    if not Process.alive?(pid) do
      IO.puts("process is not alive")
      {:error, "process not alive"}
    else
      send(pid, {self(), command, pattern})
      receive do
        result ->
          result
      after 5000 ->
          msg = "timed out"
          IO.puts(msg)
          {:error, msg}
      end
    end
  end

  @doc """
  Implements a process that handles requests for xpath queries.
  """
  @spec query_proc({pid, atom, String.t()}) :: any
  def query_proc(tree) do
    receive do
      {pid, :exit, _pattern} ->
        send(pid, :ok)
      {pid, :get_elements, pattern} ->
        elements = get_elements(tree, pattern)
        send(pid, elements)
        query_proc(tree)
      {pid, :show_tags, pattern} ->
        result = show_tags(tree, pattern)
        send(pid, result)
        query_proc(tree)
      {pid, :show_text, pattern} ->
        result = show_text(tree, pattern)
        send(pid, result)
        query_proc(tree)
      {pid, :show_namespace_defs, pattern} ->
        result = show_namespace_defs(tree, pattern)
        send(pid, result)
        query_proc(tree)
      {pid, :show_attributes, pattern} ->
        result = show_attributes(tree, pattern)
        send(pid, result)
        query_proc(tree)
      {pid, _, _} ->
        msg =  "available ops are -- :get_elements, :show_tags, " <>
          ":show_text, :show_namespace_defs, :show_attributes, :exit"
        send(pid, {:error, msg})
        query_proc(tree)
    end
  end

  @doc """
  Find and return a list of elements.
  These tuples are the Xmerl data structures.
  """
  @spec get_elements(String.t(), String.t()) :: [Tuple.t()]
  def get_elements(file_path, pattern) when is_binary(file_path) do
    tree = File.stream!(file_path) |> SweetXml.parse()
    {:ok, elements} = get_elements(tree, pattern)
    {:ok, elements}
  end
  @spec get_elements(Tuple.t(), String.t()) :: [Tuple.t()]
  def get_elements(tree, pattern) when is_tuple(tree) do
    elements = tree
      |> SweetXml.xpath(~x"//#{pattern}"l)
    {:ok, elements}
  end

  @doc """
  Show the tags of elements matching an xpath pattern.
  """
  @spec show_tags(String.t(), String.t()) :: :ok
  def show_tags(file_path, pattern) when is_binary(file_path) do
    tree = File.stream!(file_path) |> SweetXml.parse()
    show_tags(tree, pattern)
  end
  @spec show_tags(Tuple.t(), String.t()) :: :ok
  def show_tags(tree, pattern) when is_tuple(tree) do
    tree
    |> SweetXml.xpath(~x"//#{pattern}"l)
    |> Enum.each(fn (el) ->
      IO.puts("tag: #{XmerlRecs.xmlElement(el, :name)}")
    end)
    :ok
  end

  @doc """
  Show the text (content) of elements that match an xpath pattern.
  """
  @spec show_text(String.t(), String.t()) :: :ok
  def show_text(file_path, pattern) when is_binary(file_path) do
    tree = File.stream!(file_path) |> SweetXml.parse()
    show_tags(tree, pattern)
  end
  @spec show_text(Tuple.t(), String.t()) :: :ok
  def show_text(tree, pattern) when is_tuple(tree) do
    tree
    |> SweetXml.xpath(~x"//#{pattern}"l)
    |> Enum.each(fn (el) ->
      tag = XmerlRecs.xmlElement(el, :name)
      XmerlRecs.xmlElement(el, :content)
      |> Enum.each(fn (item) ->
        text = XmerlRecs.xmlText(item, :value)
        IO.puts("tag: #{tag}  text: \"#{text}\"")
      end)
    end)
    :ok
  end

  @doc """
  Show the name space defs (nsmap) of elements that match an xpath pattern.
  """
  @spec show_namespace_defs(String.t(), String.t()) :: :ok
  def show_namespace_defs(file_path, pattern) when is_binary(file_path) do
    File.stream!(file_path) |> SweetXml.parse()
    |> show_namespace_defs(pattern)
  end
  @spec show_namespace_defs(Tuple.t(), String.t()) :: :ok
  def show_namespace_defs(tree, pattern) when is_tuple(tree) do
    tree
    |> SweetXml.xpath(~x"//#{pattern}"l)
    |> Enum.each(fn (el) ->
      # extract the tag from the element record.
      tag = XmerlRecs.xmlElement(el, :name)
      IO.puts("element #{tag} -- nsmap:")
      # extract the name space record from the element record.
      ns_defs = XmerlRecs.xmlElement(el, :namespace)
      # extract the nodes (list of name space defs) from the name space record.
      # it's a list of 2-tuples: {prefix, uri}.
      XmerlRecs.xmlNamespace(ns_defs, :nodes)
      |> Enum.each(fn ({prefix, uri}) ->
        IO.puts("    prefix: #{prefix}  definition: #{uri}")
      end)
    end)
    :ok
  end

  @doc """
  Show the attributes of elements that match an xpath pattern.
  """
  @spec show_attributes(String.t(), String.t()) :: :ok
  def show_attributes(file_path, pattern) when is_binary(file_path) do
    File.stream!(file_path)
    |> SweetXml.parse()
    |> show_attributes(pattern)
  end
  @spec show_attributes(Tuple.t(), String.t()) :: :ok
  def show_attributes(tree, pattern) when is_tuple(tree) do
    tree
    |> SweetXml.xpath(~x"//#{pattern}"l)
    |> Enum.each(fn (el) ->
      # extract the tag from the element record.
      tag = XmerlRecs.xmlElement(el, :name)
      IO.puts("element #{tag} -- attributes:")
      # extract the attributes from the element record.
      XmerlRecs.xmlElement(el, :attributes)
      |> Enum.each(fn (attr) ->
        name = XmerlRecs.xmlAttribute(attr, :name)
        value = XmerlRecs.xmlAttribute(attr, :value)
        IO.puts("    name: #{name}  value: #{value}")
      end)
    end)
    :ok
  end

end
